/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <gtk/gtk.h>

#include <sndfile.h>

#include <sweep/sweep_types.h>
#include <sweep/sweep_sample.h>
#include <sweep/sweep_i18n.h>

#include "sweep_app.h"
#include "file_sndfile.h"
#include "sample.h"
#include "sample-display.h"
#include "question_dialogs.h"
#include "preferences.h"

#define LAST_LOAD_KEY "Last_Load"
#define LAST_SAVE_KEY "Last_Save"

extern sw_sample * vorbis_sample_reload (sw_sample * sample);
extern sw_sample * vorbis_sample_load (char * pathname);
extern int vorbis_sample_save (sw_sample * sample);
extern int vorbis_save_options_dialog (sw_sample * sample);

/* This source code can compile against both version 0 and version 1
** of libsndfile. The code for version 0 is at the top and the code
** for version 1 is below.
*/

extern int sndfile_save_options_dialog (sw_sample * sample);

typedef struct {
  sw_file_format_t file_format;
  gchar * name;
  gchar * exts; /* comma separated extensions */
} sw_file_format_desc;

static GtkWidget *
create_save_menu (sw_sample * sample) ;

void
file_guess_method (sw_sample * sample);

static void
file_set_format_cb (GtkWidget * widget, gpointer data) ;

#if ! (defined (SNDFILE_1))

static sw_file_format_desc file_formats [] = {
  {
    SWEEP_FILE_FORMAT_RAW,
    "Raw PCM (headerless)",
    "raw",
  },
  {
    SWEEP_FILE_FORMAT_AIFF,
    "Apple/SGI AIFF,AIFF-C",
    "aiff,aifc"
  },
  {
    SWEEP_FILE_FORMAT_AU,
    "Sun/NeXT AU",
    "au",
  },
  {
    SWEEP_FILE_FORMAT_SVX,
    "Amiga IFF,SVX8,SVX16",
    "svx",
  },
  {
    SWEEP_FILE_FORMAT_PAF,
    "Ensoniq PARIS",
    "paf",
  },
  {
    SWEEP_FILE_FORMAT_IRCAM,
    "IRCAM/Berkeley/CARL SF",
    "sf",
  },
  {
    SWEEP_FILE_FORMAT_VOC,
    "Creative Labs VOC",
    "voc",
  },
  {
    SWEEP_FILE_FORMAT_WAV,
    "Microsoft WAV",
    "wav,riff",
  },
};

sw_file_format_t
guess_file_format (gchar * pathname)
{
  int i;
  sw_file_format_desc * desc;
  sw_file_format_t ret = SWEEP_FILE_FORMAT_WAV;
  gchar ** exts, **e;
  gchar * ext;

  ext = strrchr (pathname, '.');

  if (ext != NULL) {
    ext++;

    for (i = 0; i <  sizeof(file_formats)/sizeof(sw_file_format_desc); i++) {
      desc = &file_formats[i];
      exts = g_strsplit (desc->exts, ",", 16);
      for (e = exts; *e; e++) {
	if (!g_strncasecmp (*e, ext, SW_DIR_LEN)) {
	  ret = desc->file_format;
	}
      }
      g_strfreev (exts);
    }
  }

  return ret;
}    

static GtkWidget *
create_save_menu (sw_sample * sample)
{
  GtkWidget * menu;
  GtkWidget * menuitem;
  int i;
  sw_file_format_desc * desc;
  sw_ff_instance * ffi;

  menu = gtk_menu_new ();

  sample->file_format = SWEEP_FILE_FORMAT_BY_EXTENSION;

  menuitem = gtk_menu_item_new_with_label (_("By extension"));
  gtk_menu_append (GTK_MENU(menu), menuitem);
  ffi = sw_ff_instance_new (sample, SWEEP_FILE_FORMAT_BY_EXTENSION);
  gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
		      GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new(); /* Separator */
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show(menuitem);

  for (i = 0; i < sizeof(file_formats)/sizeof(sw_file_format_desc); i++) {
    desc = &file_formats[i];
    menuitem = gtk_menu_item_new_with_label (desc->name);
    gtk_menu_append (GTK_MENU(menu), menuitem);
    ffi = sw_ff_instance_new (sample, desc->file_format);
    gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
    gtk_widget_show (menuitem);
  }

  return menu;
}

#endif 

sw_sample *
sample_reload (sw_sample * sample)
{
  sw_sample * new_sample;

  new_sample = vorbis_sample_reload (sample);

  if (new_sample == NULL)
    new_sample = sndfile_sample_reload (sample);

  return new_sample;
}

sw_sample *
sample_load (char * pathname)
{
  sw_sample * sample;

  prefs_set_string (LAST_LOAD_KEY, pathname);

  sample = vorbis_sample_load (pathname);

  if (sample == NULL)
    sample = sndfile_sample_load (pathname);

  return sample;
}

static void
sample_load_ok_cb(GtkWidget * widget, gpointer data)
{
  gchar *dir;

  dir = gtk_file_selection_get_filename(GTK_FILE_SELECTION(data));

  sample_load(dir);

  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
sample_load_cancel_cb(GtkWidget * widget, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

void
sample_load_cb(GtkWidget * widget, gpointer data)
{
  GtkWidget *filesel;
  char * load_current_file;
  gint win_width, win_height;

  win_width = gdk_screen_width () / 2;
  win_height = gdk_screen_height () / 2;

  filesel = gtk_file_selection_new(_("Sweep: Load file"));
  gtk_window_set_position (GTK_WINDOW (filesel), GTK_WIN_POS_CENTER);
  gtk_widget_set_usize (filesel, win_width, win_height);

  load_current_file = prefs_get_string (LAST_LOAD_KEY);

  if(load_current_file) {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), 
				      load_current_file);
      free (load_current_file);
  }

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		     "clicked", GTK_SIGNAL_FUNC(sample_load_ok_cb), filesel);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(sample_load_cancel_cb), filesel);

  gtk_widget_show(filesel);
}

gboolean
sample_mtime_changed (sw_sample * sample)
{
  char pathname [SW_DIR_LEN], *p;
  struct stat statbuf;

  if (*sample->directory) {
    snprintf (pathname, SW_DIR_LEN,
	      "%s/%s", sample->directory, sample->filename);
    p = pathname;
  } else {
    /* if no path has been specified, this buffer is not from disk so
     * its mtime cannot have changed */
    return FALSE;
  }

  stat (p, &statbuf);

  return (sample->last_mtime != statbuf.st_mtime);
}

void
sample_revert_ok_cb (GtkWidget * widget, gpointer data)
{
  sw_sample * sample = (sw_sample *)data;
  char pathname [SW_DIR_LEN];

  snprintf (pathname, SW_DIR_LEN,
	    "%s/%s", sample->directory,
	    sample->filename);

  sample_reload (sample);
}

void
sample_revert_cb (GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
#undef BUF_LEN
#define BUF_LEN 512
  char buf[BUF_LEN];
  char pathname [SW_DIR_LEN];
    
  sample = view->sample;

  snprintf (pathname, SW_DIR_LEN,
	    "%s/%s", sample->directory,
	    sample->filename);

  snprintf (buf, BUF_LEN,
	    _("Are you sure you want to revert %s to\n%s?\n\n"
	      "All changes and undo information will be lost."),
	    sample->filename, pathname);

  question_dialog_new (sample, _("Revert file"), buf,
		       _("Revert"), _("Don't revert"),
		       sample_revert_ok_cb, sample, NULL, NULL,
		       SWEEP_EDIT_MODE_ALLOC);
}

static void
sample_save_as_ok_cb(GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
  GtkWidget * filesel;
  gchar * pathname, * dn, * fn;

  sample = view->sample;
  filesel = gtk_widget_get_toplevel (widget);

  pathname = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));

  prefs_set_string (LAST_SAVE_KEY, pathname);

  fn = g_basename (pathname);
  dn = g_dirname (pathname);

  sample_set_pathname (sample, dn, fn);

  if (sample->file_method == SWEEP_FILE_METHOD_BY_EXTENSION) {
    file_guess_method (sample);
  }

  switch (sample->file_method) {
  case SWEEP_FILE_METHOD_LIBSNDFILE:
    sndfile_save_options_dialog (sample);
    break;
  case SWEEP_FILE_METHOD_OGGVORBIS:
    vorbis_save_options_dialog (sample);
    break;
  default:
    g_assert_not_reached ();
    break;
  }

  gtk_widget_destroy(GTK_WIDGET(filesel));
}

static void
sample_save_as_cancel_cb(GtkWidget * widget, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
file_set_format_cb (GtkWidget * widget, gpointer data)
{
  sw_sample * sample = (sw_sample *)data;

  sample->file_method = (sw_file_method_t)
    GPOINTER_TO_INT(gtk_object_get_data (GTK_OBJECT(widget), "method"));

  sample->file_format =
    GPOINTER_TO_INT(gtk_object_get_data (GTK_OBJECT(widget), "format"));
}


void
sample_save_as_cb(GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
  GtkWidget * filesel;
  GtkWidget * save_options;
  GtkWidget * frame;
  GtkWidget * hbox;
  GtkWidget * label;
  GtkWidget * option_menu;
  GtkWidget * save_menu;

  char * save_current_file;

  gint win_width, win_height;

  win_width = gdk_screen_width () / 2;
  win_height = gdk_screen_height () / 2;

  sample = view->sample;

  chdir(sample->directory);

  filesel = gtk_file_selection_new(_("Sweep: Save file"));
  gtk_window_set_position (GTK_WINDOW (filesel), GTK_WIN_POS_CENTER);
  gtk_widget_set_usize (filesel, win_width, win_height);

  save_current_file = prefs_get_string (LAST_SAVE_KEY);

  if(save_current_file) {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), 
				      save_current_file);
      g_free (save_current_file);
  }

  save_options = gtk_hbox_new (TRUE, 1);

  frame = gtk_frame_new (_("Save Options"));
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start (GTK_BOX (save_options), frame, TRUE, TRUE, 4);
  
  hbox = gtk_hbox_new (FALSE, 4);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 4);
  gtk_container_add (GTK_CONTAINER (frame), hbox);
  gtk_widget_show (hbox);
  
  label = gtk_label_new (_("Determine File Type:"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);
  
  option_menu = gtk_option_menu_new ();
  gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
  gtk_widget_show (option_menu);
  
  save_menu = create_save_menu (sample);
  gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), save_menu);
  
  gtk_widget_show (frame);

  /* pack the containing save_options hbox into the save-dialog */
  gtk_box_pack_end (GTK_BOX (GTK_FILE_SELECTION (filesel)->main_vbox),
		    save_options, FALSE, FALSE, 0);

  gtk_widget_show (save_options);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		     "clicked", GTK_SIGNAL_FUNC(sample_save_as_ok_cb), data);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(sample_save_as_cancel_cb), filesel);

  gtk_widget_show(filesel);
}

static void
sample_save_ok_cb(GtkWidget * widget, gpointer data)
{
  sw_sample * sample = (sw_sample *)data;

  sndfile_sample_save (sample);
}

void
sample_save_cb (GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
#undef BUF_LEN
#define BUF_LEN 512
  char buf[BUF_LEN];
  char pathname [SW_DIR_LEN];
    
  sample = view->sample;

  if (!sample->filename) {
    sample_save_as_cb (widget, data);
  } else {

    snprintf (pathname, SW_DIR_LEN,
	      "%s/%s", sample->directory,
	      sample->filename);
    
    snprintf (buf, BUF_LEN,
	      _("%s\n has changed on disk.\n\n"
		"Are you sure you want to save?"),
	      pathname);
    
    if (sample_mtime_changed (sample)) {
      question_dialog_new (sample, _("File modified"), buf,
			   _("Save"), _("Don't save"),
			   sample_save_ok_cb, sample, NULL, NULL,
			   SWEEP_EDIT_MODE_ALLOC);
    } else {
      sample_save_ok_cb (widget, sample);
    }
  }
}

#ifdef SNDFILE_1
/*===========================================================================================
** libsndfile-1.0.X support below. 
*/

#include <sndfile.h>

void
file_guess_method (sw_sample * sample)
{
  int i;
  gchar * ext;

  SF_FORMAT_INFO  info ;
  int	count ;

  g_assert (sample->file_method == SWEEP_FILE_METHOD_BY_EXTENSION);

  if (sample->filename == NULL) goto guess_raw;

  ext = strrchr (sample->filename, '.');

  if (ext != NULL) {
    ext++;

    if (!g_strncasecmp (ext, "ogg", SW_DIR_LEN)) {
      sample->file_method = SWEEP_FILE_METHOD_OGGVORBIS;
      sample->file_format = 0;
      return;
    }

    sf_command (NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof (int)) ;

    for (i = 0; i <  count ; i++) {
      info.format = i ;
      sf_command (NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof (info)) ;
		
      if (!g_strncasecmp (ext, info.extension, SW_DIR_LEN)) {
	sample->file_method = SWEEP_FILE_METHOD_LIBSNDFILE;
	sample->file_format = info.format;
	break; /* NB. this break is essential to ensure that eg. a file
		* ending in ".wav" is guessed as RIFF/WAV not Sphere/NIST,
		* which also may use the ending ".wav"; assuming RIFF/WAV is
		* more common we prefer to use this format when guessing.
		*/
      }
    }

    /* Catch some common conventions not offered by libsndfile */
    if (sample->file_method == SWEEP_FILE_METHOD_BY_EXTENSION) {
      if (!g_strncasecmp(ext, "aifc", SW_DIR_LEN) ||
	  !g_strncasecmp(ext, "aif", SW_DIR_LEN)) {
	sample->file_method = SWEEP_FILE_METHOD_LIBSNDFILE;
	sample->file_format = SF_FORMAT_AIFF;
      }
    }
  }

 guess_raw:
  
  if (sample->file_method == SWEEP_FILE_METHOD_BY_EXTENSION) {
    sample->file_method = SWEEP_FILE_METHOD_LIBSNDFILE;
    sample->file_format = SF_FORMAT_RAW;
  }

  return;
}    

static GtkWidget *
create_save_menu (sw_sample * sample)
{
  GtkWidget * menu;
  GtkWidget * menuitem;
  int i;

  SF_FORMAT_INFO  info ;
  int	count ;

  menu = gtk_menu_new ();

  sample->file_method = SWEEP_FILE_METHOD_BY_EXTENSION;

  menuitem = gtk_menu_item_new_with_label (_("By extension"));
  gtk_menu_append (GTK_MENU(menu), menuitem);
  gtk_object_set_data (GTK_OBJECT(menuitem), "method",
		       SWEEP_FILE_METHOD_BY_EXTENSION);
  gtk_object_set_data (GTK_OBJECT(menuitem), "format", 0);
  gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
		      GTK_SIGNAL_FUNC(file_set_format_cb), sample);
  gtk_widget_show (menuitem);

  /* Ogg Vorbis */

  menuitem = gtk_menu_item_new (); /* Separator */
  gtk_menu_append (GTK_MENU(menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Ogg Vorbis (Xiph.org)");
  gtk_menu_append (GTK_MENU(menu), menuitem);
  gtk_object_set_data (GTK_OBJECT(menuitem), "method",
		       GINT_TO_POINTER(SWEEP_FILE_METHOD_OGGVORBIS));
  gtk_object_set_data (GTK_OBJECT(menuitem), "format",
		       GINT_TO_POINTER(0));
  gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
		      GTK_SIGNAL_FUNC(file_set_format_cb), sample);
  gtk_widget_show (menuitem);

  /* libsndfile formats */

  menuitem = gtk_menu_item_new(); /* Separator */
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show(menuitem);

  sf_command (NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof (int)) ;

  for (i = 0; i < count ; i++) {
    info.format = i ;
    sf_command (NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof (info)) ;
    
    menuitem = gtk_menu_item_new_with_label (info.name);
    gtk_menu_append (GTK_MENU(menu), menuitem);
    gtk_object_set_data (GTK_OBJECT(menuitem), "method",
			 GINT_TO_POINTER(SWEEP_FILE_METHOD_LIBSNDFILE));
    gtk_object_set_data (GTK_OBJECT(menuitem), "format",
			 GINT_TO_POINTER(info.format));
    gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(file_set_format_cb), sample);
    gtk_widget_show (menuitem);
  }

  return menu;
}


#endif
